﻿using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Configuration;
using System.Linq;
using System.ServiceModel;
using IMSES_Template.ImsEnterpriseServicePort;
using IMSES_Template.Model;
using IMSES_Template.Properties;
using IMSES_Template.Utils;

namespace IMSES_Template.DataAccess
{
    public sealed class ImsEsClient : IClient
    {
        private readonly PersonManagementServiceSyncClient _personService;
        private readonly GroupManagementServiceSyncClient _groupService;
        private readonly MembershipManagementServiceSyncClient _memberService;
        private readonly ExtendedPersonManagementServiceSyncClient _extendedService;

        public string MessageIdentifier;

        private const int MaxOperations = 50;

        #region Creating and Disposing

        /// <summary>
        /// Creates a new ImsEsClient, connecting the the Production endpoint.
        /// </summary>
        /// <param name="userName">Username needed for ImsEs.</param>
        /// <param name="passWord">Password needed for ImsEs.</param>
        public ImsEsClient(string userName, string passWord) 
            : this(userName, passWord, ItslearningServers.Production)
        {
        }

        /// <summary>
        /// Creates a new ImsEsClient, connecting to the given server.
        /// </summary>
        /// <param name="userName">Username needed for ImsEs.</param>
        /// <param name="passWord">Password needed for ImsEs.</param>
        /// <param name="server">Server to connect to.</param>
        public ImsEsClient(string userName, string passWord, ItslearningServers server)
            : this(userName, passWord, GetEndpoint(server))
        {
        }

        private static string GetEndpoint(ItslearningServers server)
        {
            var settingsProperty = Settings.Default[server.ToString()];
            return settingsProperty != null ? settingsProperty.ToString() : Settings.Default.Production;
        }

        /// <summary>
        /// Opens a new ImsEsClient, using the given endpoint, username and password.
        /// </summary>
        /// <param name="userName">Username for ImsEs.</param>
        /// <param name="passWord">Password for EmsEs.</param>
        /// <param name="endpointAddress">Address of the ImsEnterpriseServicesPort.svc file in Itslearning.</param>
        public ImsEsClient(string userName, string passWord, string endpointAddress)
        {
            // Check the binding:
            BasicHttpBinding binding;
            try
            {
                binding = new BasicHttpBinding("ItslearningImsEsConfigBinding");
            }
            catch (Exception ex)
            {
                if (ex is ConfigurationErrorsException || ex is KeyNotFoundException)
                {
                    binding = new BasicHttpBinding(BasicHttpSecurityMode.TransportWithMessageCredential);
                }
                else
                {
                    throw;
                }
            }

            _personService = new PersonManagementServiceSyncClient(binding, new EndpointAddress(endpointAddress));
            _groupService = new GroupManagementServiceSyncClient(binding, new EndpointAddress(endpointAddress));
            _memberService = new MembershipManagementServiceSyncClient(binding, new EndpointAddress(endpointAddress));
            _extendedService = new ExtendedPersonManagementServiceSyncClient(binding, new EndpointAddress(endpointAddress));

            OpenClient(_personService, userName, passWord);
            OpenClient(_groupService, userName, passWord);
            OpenClient(_memberService, userName, passWord);
            OpenClient(_extendedService, userName, passWord);
        }


        /// <summary>
        /// Opens the specified service, after entering the credentials.
        /// </summary>
        /// <typeparam name="T">class of the service.</typeparam>
        /// <param name="service">Service to be opened.</param>
        /// <param name="username">Username for the service.</param>
        /// <param name="password">Password for the service.</param>
        private void OpenClient<T>(ClientBase<T> service, string username, string password) where T : class
        {
            if (service.State == CommunicationState.Opened) return;

            // Fill the credentials and open the client.
            // ReSharper disable PossibleNullReferenceException
            service.ClientCredentials.UserName.UserName = username;
            service.ClientCredentials.UserName.Password = password;
            // ReSharper restore PossibleNullReferenceException

            service.Open();
        }


        /// <summary>
        /// When closing the object the ImsEs clients also have to be closed.
        /// </summary>
        public void Dispose()
        {
            if (_personService.State != CommunicationState.Faulted)     { _personService.Close(); }
            if (_groupService.State != CommunicationState.Faulted)      { _groupService.Close(); }
            if (_memberService.State != CommunicationState.Faulted)     { _memberService.Close(); }
            if (_extendedService.State != CommunicationState.Faulted)   { _extendedService.Close(); }
        }

        #endregion
        
        
        #region Errors and Warnings

        /// <summary>
        /// Checks the given HeaderInfo for errors and warnings, if any error appears throws an exception, and returns the warnings.
        /// </summary>
        /// <param name="responseInfo">The Header info from the response.</param>
        /// <param name="identifier">The object, for using as identifier in the result.</param>
        /// <returns>The warning.</returns>
        private static OperationResult CheckSingleErrorWarning(syncResponseHeaderInfo responseInfo, string identifier)
        {
            // First check for an exception.
            CheckError(responseInfo, identifier);

            return CheckWarning(responseInfo, identifier);
        }

        /// <summary>
        /// Checks the Header info for errors, and returns this error.
        /// </summary>
        /// <param name="responseInfo">The HeaderInfo from the response from Itslearning.</param>
        /// <param name="identifier">The object, for using as identifier in the result.</param>
        /// <returns>The error.</returns>
        private static void CheckError(syncResponseHeaderInfo responseInfo, string identifier)
        {
            // Because there are two possible response formats, first it must be checked which one is given.
            var info = responseInfo.statusInfo ?? responseInfo.statusInfoSet[0];

            if (info.severity != StatusInfoDTypeSeverity.error) return;

            var error = info.description.text.Length > 0 ? info.description.text : info.codeMinor[0].codeMinorValue;
            var result = new OperationResult(ResultStatus.Error, identifier, error);
            
            // And throw the exception.
            throw new ImsEsException(result);
        }

        /// <summary>
        /// Checks the Header info for warnings, and returns this warning.
        /// </summary>
        /// <param name="responseInfo">The HeaderInfo from the response from Itslearning.</param>
        /// <param name="returnObject">The object from the request.</param>
        /// <returns>The warning.</returns>
        private static OperationResult CheckWarning(syncResponseHeaderInfo responseInfo, string returnObject)
        {
            // Because there are two possible response formats, first it must be checked which one is given.
            var info = responseInfo.statusInfo ?? responseInfo.statusInfoSet[0];

            if (info.severity == StatusInfoDTypeSeverity.warning)
            {
                return new OperationResult(ResultStatus.Warning, returnObject, info.description.text);
            }

            return new OperationResult(ResultStatus.Success, returnObject, null);
        }



        /// <summary>
        /// Checks the given HeaderInfo for errors and warnings, if any error appears throws an exception, and returnes the warnings.
        /// </summary>
        /// <param name="responseInfo">The Header info from the response.</param>
        /// <param name="identifierList">An array with the objects, for using as key in the lists.</param>
        /// <returns>The list with warnings.</returns>
        private static BatchOperationResult CheckMultipleErrorWarning(IList<StatusInfoDType> responseInfo, IList<string> identifierList)
        {
            var errorList = CheckErrors(responseInfo, identifierList);
            var warningList = CheckWarnings(responseInfo, identifierList);

            if (errorList.Count > 0)
            {
                var response = new BatchOperationResult(ResultStatus.Error, errorList, warningList);

                throw new ImsEsBatchException("There were errors executing your request, please check the attached Error Messages!", response);
            }
            if (warningList.Count > 0)
            {
                return new BatchOperationResult(ResultStatus.Warning, null, warningList);
            }
            return new BatchOperationResult(ResultStatus.Success, null, null);
        }

        /// <summary>
        /// Checks the Header info for errors, and returns these errors.
        /// </summary>
        /// <param name="responseInfo">The HeaderInfo from the response from Itslearning.</param>
        /// <param name="returnObjectList">An array with the objects from the request.</param>
        /// <returns>All errors.</returns>
        private static List<OperationResult> CheckErrors(IList<StatusInfoDType> responseInfo, IList<string> returnObjectList)
        {
            return (from info in responseInfo 
                    let index = responseInfo.IndexOf(info) 
                    where info.severity == StatusInfoDTypeSeverity.error 
                    let errorText = info.description.text.Length > 0 ? info.description.text : info.codeMinor[0].codeMinorValue 
                    select new OperationResult(ResultStatus.Error, returnObjectList[index], errorText))
                    .ToList();
        }

        /// <summary>
        /// Checks the Header info for warnings, and returns these warnings.
        /// </summary>
        /// <param name="responseInfo">The HeaderInfo from the response from Itslearning.</param>
        /// <param name="returnObjectList">An array with the objects from the request.</param>
        /// <returns>All warnings.</returns>
        private static List<OperationResult> CheckWarnings(IList<StatusInfoDType> responseInfo, IList<string> returnObjectList)
        {
            return (from info in responseInfo 
                    let index = responseInfo.IndexOf(info) 
                    where info.severity == StatusInfoDTypeSeverity.warning 
                    select new OperationResult(ResultStatus.Warning, returnObjectList[index], info.description.text))
                    .ToList();
        }

        #endregion

        

        #region User - Single Functions

        /// <summary>
        /// Creates a Request header with the Message Identifier, and the given function / identifier.
        /// </summary>
        /// <param name="function">The name of the function.</param>
        /// <param name="identifier">the extra identidier.</param>
        /// <returns>The new Reuest header.</returns>
        private syncRequestHeaderInfo CreateHeader(string function, string identifier)
        {
            return new syncRequestHeaderInfo
            {
                messageIdentifier = string.Format("{0} - {1} {2}", MessageIdentifier, function, identifier)
            };
        }

        /// <summary>
        /// Creates a new User in Itslearning.
        /// </summary>
        /// <param name="newUser">The new user to be created.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the new User is not valid.</exception>
        public OperationResult CreateUser(User newUser)
        {
            var headerInfo = CreateHeader("CreateUser", newUser.SyncKey); 
            var createRequest = new createPersonRequest
            {
                sourcedId = new createPersonRequestSourcedId { identifier = newUser.SyncKey },
                person = newUser.ToImsEsPerson(true)
            };
            createPersonResponse createResponse;
            
            // Sends the user to Itslearning.
            var response = _personService.createPerson(headerInfo, createRequest, out createResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, newUser.SyncKey);
        }

        /// <summary>
        /// Retrieves the specified User from Itslearning.
        /// </summary>
        /// <param name="syncKey">The specified User key.</param>
        /// <param name="readUser">Out: The retrieved User.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult ReadUser(string syncKey, out User readUser)
        {
            var headerInfo = CreateHeader("ReadUser", syncKey);
            var readRequest = new readPersonRequest { sourcedId = new readPersonRequestSourcedId { identifier = syncKey } };
            readPersonResponse readResponse;

            // Get the user from Itslearning.
            var response = _personService.readPerson(headerInfo, readRequest, out readResponse);
            readUser = readResponse.person.ToItsUser(syncKey);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, syncKey);
        }

        /// <summary>
        /// Updates an User in Itslearning.
        /// </summary>
        /// <param name="updatedUser">The updated user.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the updated User is not valid.</exception>
        public OperationResult UpdateUser(User updatedUser)
        {
            var headerInfo = CreateHeader("UpdateUser", updatedUser.SyncKey);
            var updateRequest = new updatePersonRequest
                {
                    sourcedId = new updatePersonRequestSourcedId {identifier = updatedUser.SyncKey},
                    person = updatedUser.ToImsEsPerson(false)
                };
            updatePersonResponse updateResponse;

            // Sends the updated user to Itslearning.
            var response = _personService.updatePerson(headerInfo, updateRequest, out updateResponse);
            
            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, updatedUser.SyncKey);
        }

        /// <summary>
        /// Replaces an User in Itslearning (Update & empties all fields that are not changed).
        /// </summary>
        /// <param name="updatedUser">The updated user.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the replaced User is not valid.</exception>
        public OperationResult ReplaceUser(User updatedUser)
        {
            var headerInfo = CreateHeader("ReplaceUser", updatedUser.SyncKey);
            var replaceRequest = new replacePersonRequest
                {
                    sourcedId = new replacePersonRequestSourcedId {identifier = updatedUser.SyncKey},
                    person = updatedUser.ToImsEsPerson(true)
                };
            replacePersonResponse replaceResponse;

            // Sends the updated user to Itslearning.
            var response = _personService.replacePerson(headerInfo, replaceRequest, out replaceResponse);
            
            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, updatedUser.SyncKey);
        }

        /// <summary>
        /// Deletes the specified User in Itslearning.
        /// </summary>
        /// <param name="syncKey">The key that must be deleted.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult DeleteUser(string syncKey)
        {
            var headerInfo = CreateHeader("DeleteUser", syncKey);
            var deleteRequest = new deletePersonRequest
            {
                sourcedId = new deletePersonRequestSourcedId { identifier = syncKey },
            };
            deletePersonResponse deleteResponse;

            // Deletes the person from Itslearning.
            var response = _personService.deletePerson(headerInfo, deleteRequest, out deleteResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, syncKey);
        }

        /// <summary>
        /// Changes the SyncKey of an User in Itslearning.
        /// </summary>
        /// <param name="oldKey">The old SyncKey.</param>
        /// <param name="newKey">The new SyncKey.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult ChangeUserKey(string oldKey, string newKey)
        {
            var headerInfo = CreateHeader("ChangeUserKey", oldKey);
            var changeRequest = new changePersonIdentifierRequest
            {
                sourcedId = new changePersonIdentifierRequestSourcedId { identifier = oldKey },
                newSourcedId = new changePersonIdentifierRequestNewSourcedId { identifier = newKey }
            };
            changePersonIdentifierResponse changeResponse;

            // Sends the new key to Itslearning.
            var response = _personService.changePersonIdentifier(headerInfo, changeRequest, out changeResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, oldKey);
        }

        /// <summary>
        /// Retrieves all Users from Itslearning who are added to the given Hierarchy.
        /// </summary>
        /// <param name="syncId">The Hierarchy.</param>
        /// <param name="readUserList">Out: The list with Users.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult ReadUsersFromHierarchy(string syncId, out List<User> readUserList)
        {
            var headerInfo = CreateHeader("ReadUserFromHierarchy", syncId);
            var readRequest = new readPersonsForGroupRequest { groupSourcedId = new readPersonsForGroupRequestGroupSourcedId{ identifier = syncId } };
            readPersonsForGroupResponse readResponse;

            // Get the users from Itslearning.
            var response = _personService.readPersonsForGroup(headerInfo, readRequest, out readResponse);
            readUserList = readResponse.personIdPairSet.ToItsUserList();

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, syncId);
        }

        /// <summary>
        /// Retrieves all Users from Itslearning.
        /// </summary>
        /// <param name="readUserList">Out: The list with Users.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult ReadAllUsers(out List<User> readUserList)
        {
            var headerInfo = CreateHeader("ReadAllUsers","");
            var readRequest = new readAllPersonsRequest();
            readAllPersonsResponse readResponse;

            // Get all users from Itslearning.
            var response = _extendedService.readAllPersons(headerInfo, readRequest, out readResponse);
            readUserList = readResponse.personIdPairSet.ToItsUserList();

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, "");
        }

        #endregion


        #region User - Batch Functions

        /// <summary>
        /// Creates the specified Users in Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="newUsers">A collection with the new Users.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If one of the new Users is not valid.</exception>
        public BatchOperationResult CreateUsers(ICollection<User> newUsers)
        {
            var maxCount = newUsers.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncKeys = new List<string>(maxCount);
            var objectSet = new List<PersonIdPairDType> (MaxOperations);
            
            foreach (var newUser in newUsers)
            {
                objectSet.Add(new PersonIdPairDType
                    {
                        person = newUser.ToImsEsPerson(true),
                        sourcedId = new PersonIdPairDTypeSourcedId{ identifier = newUser.SyncKey }
                    });
                syncKeys.Add(newUser.SyncKey);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    CreateUsers(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                CreateUsers(objectSet, ref statusInfo, index, maxIndex);
            }


            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncKeys);
        }

        /// <summary>
        /// Private function to send a CreateUsersBatch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The persons.</param>
        /// <param name="statusInfo">Statusinfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void CreateUsers(List<PersonIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("CreateMultipleUsers", string.Format("{0}v{1}", index, maxIndex));
            var createRequest = new createPersonsRequest { personIdPairSet = objectSet.ToArray() };
            createPersonsResponse createResponse;

            // Send the users to Itslearning:
            var response = _personService.createPersons(headerInfo, createRequest, out createResponse);
            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Retrieves the specified Users from Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="readSyncKeys">Collection with users to be retrieved.</param>
        /// <param name="readUserList">Out: The list with Users.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        public BatchOperationResult ReadUsers(ICollection<string> readSyncKeys, out List<User> readUserList)
        {
            var maxCount = readSyncKeys.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            readUserList = new List<User>();
            var statusInfo = new List<StatusInfoDType>();
            var syncKeys = new List<string>(maxCount);
            var objectSet = new List<string>();
            
            foreach (var syncKey in readSyncKeys)
            {
                objectSet.Add(syncKey);
                syncKeys.Add(syncKey);

                // If enough objects are added send them to Itslearning.
                if (objectSet.Count >= MaxOperations )
                {
                    ReadUsers(objectSet, ref statusInfo, index, maxIndex, ref readUserList);
                    
                    objectSet.Clear();
                    index ++;
                }
            }
            // And finally send the last batch to Itslearning.
            if (objectSet.Count > 0)
            {
                ReadUsers(objectSet, ref statusInfo, index, maxIndex, ref readUserList);
            }
            
            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncKeys);
        }

        /// <summary>
        /// Private function to read the users from Itslearning.
        /// </summary>
        /// <param name="objectSet">List with users.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        /// <param name="readUserList">The list with Users, that is filled.</param>
        private void ReadUsers(List<string> objectSet, ref List<StatusInfoDType> statusInfo,
                                                       int index, int maxIndex, ref List<User> readUserList)
        {
            var headerInfo = CreateHeader("ReadMultipleUsers", string.Format("{0}-{1}", index, maxIndex));
            var readRequest = new readPersonsRequest {sourcedIdSet = objectSet.ToArray()};
            readPersonsResponse readResponse;

            // Retrieve the users from Itslearning.
            var response = _personService.readPersons(headerInfo, readRequest, out readResponse);
            readUserList.AddRange(readResponse.personSet.ToItsUserList(objectSet, response));

            statusInfo.AddRange(response.statusInfoSet ?? new[] {response.statusInfo});
        }


        /// <summary>
        /// Updates the given Users in Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="updatedUsers">The updated Users.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If one of the updated Users is not valid.</exception>
        public BatchOperationResult UpdateUsers(ICollection<User> updatedUsers)
        {
            var maxCount = updatedUsers.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncKeys = new List<string>(maxCount);
            var objectSet = new List<PersonIdPairDType>(MaxOperations);

            foreach (var updatedUser in updatedUsers)
            {
                objectSet.Add(new PersonIdPairDType
                {
                    person = updatedUser.ToImsEsPerson(false),
                    sourcedId = new PersonIdPairDTypeSourcedId { identifier = updatedUser.SyncKey }
                });
                syncKeys.Add(updatedUser.SyncKey);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    UpdateUsers(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                UpdateUsers(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncKeys);
        }

        /// <summary>
        /// Private function to send a UpdateUsers Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The persons.</param>
        /// <param name="statusInfo">StatusInfo</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void UpdateUsers(List<PersonIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("UpdateMultipleUsers", string.Format("{0}-{1}", index, maxIndex));

            var updateRequest = new updatePersonsRequest {personIdPairSet = objectSet.ToArray()};
            updatePersonsResponse updateResponse;

            // Sends the updated users to Itslearning.
            var response = _personService.updatePersons(headerInfo, updateRequest, out updateResponse);
            
            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Replaces the given Users in Itslearning (Update + all fields that are not given are emptied).
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="updatedUsers">The updated Users.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If one of the replaced Users is not valid.</exception>
        public BatchOperationResult ReplaceUsers(ICollection<User> updatedUsers)
        {
            var maxCount = updatedUsers.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncKeys = new List<string>(maxCount);
            var objectSet = new List<PersonIdPairDType>(MaxOperations);

            foreach (var updatedUser in updatedUsers)
            {
                objectSet.Add(new PersonIdPairDType
                    {
                        person = updatedUser.ToImsEsPerson(true),
                        sourcedId = new PersonIdPairDTypeSourcedId {identifier = updatedUser.SyncKey}
                    });
                syncKeys.Add(updatedUser.SyncKey);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    ReplaceUsers(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                ReplaceUsers(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncKeys);
        }

        /// <summary>
        /// Private function to send a ReplaceUsers batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The persons.</param>
        /// <param name="statusInfo">StatusInfo</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void ReplaceUsers(List<PersonIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("ReplaceMultipleUsers", string.Format("{0}-{1}", index, maxIndex));

            var replaceRequest = new replacePersonsRequest { personIdPairSet = objectSet.ToArray() };
            replacePersonsResponse replaceResponse;

            // Sends the updated users to Itslearning.
            var response = _personService.replacePersons(headerInfo, replaceRequest, out replaceResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Deletes the specified Users in Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="deleteSyncKeys">The Users to be deleted.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        public BatchOperationResult DeleteUsers(ICollection<string> deleteSyncKeys)
        {
            var maxCount = deleteSyncKeys.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncKeys = new List<string>(maxCount);
            var objectSet = new List<string>(MaxOperations);

            foreach (var syncKey in deleteSyncKeys)
            {
                objectSet.Add(syncKey);
                syncKeys.Add(syncKey);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    DeleteUsers(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                DeleteUsers(objectSet, ref statusInfo, index, maxIndex);
            }
            
            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncKeys);
        }

        /// <summary>
        /// Private function to send a DeleteUsers batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The persons.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void DeleteUsers(List<string> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("DeleteMultipleUsers", string.Format("{0}-{1}", index, maxIndex));
            var deleteRequest = new deletePersonsRequest { sourcedIdSet = objectSet.ToArray() };
            deletePersonsResponse deleteResponse;

            // Deletes the users from Itslearning.
            var response = _personService.deletePersons(headerInfo, deleteRequest, out deleteResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Changes the specified SyncKeys for Users in Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="syncKeyPairs">The synckeys and the new values.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        public BatchOperationResult ChangeUserKeys(ICollection<KeyPair> syncKeyPairs)
        {
            var maxCount = syncKeyPairs.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncKeys = new List<string>(maxCount);
            var objectSet = new List<IdentifierPairDType>(MaxOperations);

            foreach (var syncKeyPair in syncKeyPairs)
            {
                objectSet.Add(new IdentifierPairDType
                    {
                        firstId = syncKeyPair.OldKey,
                        secondId = syncKeyPair.NewKey
                    });
                syncKeys.Add(syncKeyPair.OldKey);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    ChangeUserKeys(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                ChangeUserKeys(objectSet, ref statusInfo, index, maxIndex);
            }
            
            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncKeys);
        }

        /// <summary>
        /// Private function to send a ChangeUserKeys batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The persons.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void ChangeUserKeys(List<IdentifierPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("ChangeMultipleUserKeys", string.Format("{0}-{1}", index, maxIndex));
            var changeRequest = new changePersonsIdentifierRequest { pairSourcedIdSet = objectSet.ToArray() };
            changePersonsIdentifierResponse changeResponse;

            // Sends the new keys to Itslearning.
            var response = _personService.changePersonsIdentifier(headerInfo, changeRequest, out changeResponse);

            statusInfo.AddRange(response.statusInfoSet ?? new[] { response.statusInfo });
        }

        #endregion


        #region Hierarchy - Single Functions

        /// <summary>
        /// Creates a new Hierarchy in Itslearning.
        /// </summary>
        /// <param name="newHierarchy">The new Hierarchy.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the new Hierarchy is not valid.</exception>
        public OperationResult CreateHierarchy(Hierarchy newHierarchy)
        {
            var headerInfo = CreateHeader("CreateHierarchy", newHierarchy.SyncId);
            var createRequest = new createGroupRequest
            {
                sourcedId = new createGroupRequestSourcedId { identifier = newHierarchy.SyncId },
                group = newHierarchy.ToImsEsGroup(true)
            };
            createGroupResponse createResponse;

            // Sends the hierarchy to Itslearning.
            var response = _groupService.createGroup(headerInfo, createRequest, out createResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, newHierarchy.SyncId);
        }

        /// <summary>
        /// Retrieves the specified Hierarchy from Itslearning.
        /// </summary>
        /// <param name="syncId">The specified Hierarchy Id.</param>
        /// <param name="readHierarchy">Out: The retrieved Hierarchy.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult ReadHierarchy(string syncId, out Hierarchy readHierarchy)
        {
            var headerInfo = CreateHeader("ReadHierarchy", syncId);
            var readRequest = new readGroupRequest
            {
                sourcedId = new readGroupRequestSourcedId { identifier = syncId },
            };
            readGroupResponse readResponse;

            // Retrieves the hierarchy from Itslearning.
            var response = _groupService.readGroup(headerInfo, readRequest, out readResponse);
            readHierarchy = readResponse.group.ToItsHierarchy(syncId);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, syncId);
        }

        /// <summary>
        /// Updates an Hierarchy in Istlearning.
        /// Note: if there is a Course attached to this Hierarchy, it is NOT updated.
        /// </summary>
        /// <param name="updatedHierarchy">The updated Hierarchy.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the updated Hierarchy is not valid.</exception>
        public OperationResult UpdateHierarchy(Hierarchy updatedHierarchy)
        {
            var headerInfo = CreateHeader("UpdateHierarchy", updatedHierarchy.SyncId);
            var updateRequest = new updateGroupRequest
                {
                    sourcedId = new updateGroupRequestSourcedId {identifier = updatedHierarchy.SyncId},
                    group = updatedHierarchy.ToImsEsGroup(false)
                };
            updateGroupResponse updateResponse;

            // Sends the updated hierarchy to Itslearning.
            var response = _groupService.updateGroup(headerInfo, updateRequest, out updateResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, updatedHierarchy.SyncId);
        }

        /// <summary>
        /// Replaces an Hierarchy in Istlearning (Update + all unspecified fields are emptied.).
        /// Note: if there is a Course attached to this Hierarchy, it is NOT replaced.
        /// </summary>
        /// <param name="updatedHierarchy">The updated Hierarchy.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the replaced Hierarchy is not valid.</exception>
        public OperationResult ReplaceHierarchy(Hierarchy updatedHierarchy)
        {
            var headerInfo = CreateHeader("ReplaceHierarchy", updatedHierarchy.SyncId);
            var replaceRequest = new replaceGroupRequest
                {
                    sourcedId = new replaceGroupRequestSourcedId {identifier = updatedHierarchy.SyncId},
                    group = updatedHierarchy.ToImsEsGroup(true)
                };
            replaceGroupResponse replaceResponse;

            // Sends the updated hierarchy to Itslearning.
            var response = _groupService.replaceGroup(headerInfo, replaceRequest, out replaceResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, updatedHierarchy.SyncId);
        }

        /// <summary>
        /// Deletes an Hierarchy in Itslearning.
        /// </summary>
        /// <param name="syncId">The id that must be deleted.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult DeleteHierarchy(string syncId)
        {
            var headerInfo = CreateHeader("DeleteHierarchy", syncId);
            var deleteRequest = new deleteGroupRequest
            {
                sourcedId = new deleteGroupRequestSourcedId { identifier = syncId },
            };
            deleteGroupResponse deleteResponse;

            // deletes the hierarchy in Itslearning.
            var response = _groupService.deleteGroup(headerInfo, deleteRequest, out deleteResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, syncId);
        }

        /// <summary>
        /// Changes the id for the Hierarchy in Itslearning.
        /// </summary>
        /// <param name="oldId">The old Id.</param>
        /// <param name="newId">The new Id.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult ChangeHierarchyId(string oldId, string newId)
        {
            var headerInfo = CreateHeader("ChangeHierarchyId", oldId);
            var changeRequest = new changeGroupIdentifierRequest
            {
                sourcedId = new changeGroupIdentifierRequestSourcedId { identifier = oldId },
                newSourcedId = new changeGroupIdentifierRequestNewSourcedId { identifier = newId }
            };
            changeGroupIdentifierResponse changeResponse;

            // Sends the new id to Itslearning.
            var response = _groupService.changeGroupIdentifier(headerInfo, changeRequest, out changeResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, oldId);
        }

        /// <summary>
        /// Retrieves all Hierarchies from Itslearning attached to the given User.
        /// </summary>
        /// <param name="syncKey">The User.</param>
        /// <param name="readHierarchyList">Out: The list with Hierarchies.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult ReadHierarchiesFromUser(string syncKey, out List<Hierarchy> readHierarchyList)
        {
            var headerInfo = CreateHeader("ReadHierarchiesFromUser", syncKey);
            var readRequest = new readGroupsForPersonRequest
            {
                personSourcedId = new readGroupsForPersonRequestPersonSourcedId { identifier = syncKey },
            };
            readGroupsForPersonResponse readResponse;

            // Retrieves the hierarchies from Itslearning.
            var response = _groupService.readGroupsForPerson(headerInfo, readRequest, out readResponse);
            readHierarchyList = readResponse.groupSet.ToItsHierarchyList();

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, syncKey);
        }

        #endregion


        #region Hierarchy - Batch Functions

        /// <summary>
        /// Creates the given Hierarchies in Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="newHierarchies">The new Hierarchies.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If one of the new Hierarchies is not valid.</exception>
        public BatchOperationResult CreateHierarchies(ICollection<Hierarchy> newHierarchies)
        {
            var maxCount = newHierarchies.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<GroupIdPairDType>(MaxOperations);

            foreach (var newHierarchy in newHierarchies)
            {
                objectSet.Add(new GroupIdPairDType
                    {
                        group = newHierarchy.ToImsEsGroup(true),
                        sourcedId = new GroupIdPairDTypeSourcedId{ identifier = newHierarchy.SyncId }
                    });
                syncIds.Add(newHierarchy.SyncId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    CreateHierarchies(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                CreateHierarchies(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a CreateHierarchies batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The hierarchies.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void CreateHierarchies(List<GroupIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("CreateMultipleHierarchies", string.Format("{0}-{1}", index, maxIndex));
            var createRequest = new createGroupsRequest {groupIdPairSet = objectSet.ToArray()};
            createGroupsResponse createResponse;

            // Sends the hierarchies to Itslearning.
            var response = _groupService.createGroups(headerInfo, createRequest, out createResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Retrieves the given Hierarchies from Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="readSyncIds">The given Hierarchies.</param>
        /// <param name="readHierarchyList">Out: The retrieved Hierarchies.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        public BatchOperationResult ReadHierarchies(ICollection<string> readSyncIds, out List<Hierarchy> readHierarchyList)
        {
            var maxCount = readSyncIds.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            readHierarchyList = new List<Hierarchy>();
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<string>(MaxOperations);

            foreach (var syncId in readSyncIds)
            {
                objectSet.Add(syncId);
                syncIds.Add(syncId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    ReadHierarchies(objectSet, ref statusInfo, index, maxIndex, ref readHierarchyList);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                ReadHierarchies(objectSet, ref statusInfo, index, maxIndex, ref readHierarchyList);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a ReadHierarchies Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The hierarchies.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        /// <param name="readHierarchyList">The list with retrieved Hierarchies.</param>
        private void ReadHierarchies(List<string> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex, ref List<Hierarchy> readHierarchyList)
        {
            var headerInfo = CreateHeader("ReadMultipleHierarchies", string.Format("{0}-{1}", index, maxIndex));
            var readRequest = new readGroupsRequest { sourcedIdSet = objectSet.ToArray() };
            readGroupsResponse readResponse;

            // Retrieves the hierarchies from Itslearning.
            var response = _groupService.readGroups(headerInfo, readRequest, out readResponse);
            readHierarchyList.AddRange(readResponse.groupSet.ToItsHierarchyList(objectSet, response));

            statusInfo.AddRange(response.statusInfoSet ?? new[] { response.statusInfo });
        }

        /// <summary>
        /// Updates the given Hierarchies in Itslearning.
        /// Note: if there is a Course attached to one of the Hierarchies, it is NOT updated.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="updatedHierarchies">The Hierarchies to be updated.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If one of the updated Hierarchies is not valid.</exception>
        public BatchOperationResult UpdateHierarchies(ICollection<Hierarchy> updatedHierarchies)
        {
            var maxCount = updatedHierarchies.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<GroupIdPairDType>(MaxOperations);

            foreach (var updatedHierarchy in updatedHierarchies)
            {
                objectSet.Add(new GroupIdPairDType
                    {
                        group = updatedHierarchy.ToImsEsGroup(false),
                        sourcedId = new GroupIdPairDTypeSourcedId {identifier = updatedHierarchy.SyncId}
                    });
                syncIds.Add(updatedHierarchy.SyncId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    UpdateHierarchies(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                UpdateHierarchies(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a UpdateHierarchies Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The hierarchies.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void UpdateHierarchies(List<GroupIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("UpdateMultipleHierarchies", string.Format("{0}-{1}", index, maxIndex));
            var updateRequest = new updateGroupsRequest { groupIdPairSet = objectSet.ToArray() };
            updateGroupsResponse updateResponse;

            // Sends the updated hierarchies to Itslearning.
            var response = _groupService.updateGroups(headerInfo, updateRequest, out updateResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Replaces the given Hierarchies in Itslearning (Update + any unspecified fields are emptied).
        /// (Note: if there is a Course attached to one of the Hierarchy, it is NOT replaced.)
        /// (Also: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="updatedHierarchies">The Hierarchies to be updated.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If one of the replaced Hierarchies is not valid.</exception>
        public BatchOperationResult ReplaceHierarchies(ICollection<Hierarchy> updatedHierarchies)
        {
            var maxCount = updatedHierarchies.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<GroupIdPairDType>(MaxOperations);

            foreach (var updatedHierarchy in updatedHierarchies)
            {
                objectSet.Add(new GroupIdPairDType
                    {
                        group = updatedHierarchy.ToImsEsGroup(true),
                        sourcedId = new GroupIdPairDTypeSourcedId {identifier = updatedHierarchy.SyncId}
                    });
                syncIds.Add(updatedHierarchy.SyncId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    ReplaceHierarchies(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                ReplaceHierarchies(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a ReplaceHierarchies Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The hierarchies.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void ReplaceHierarchies(List<GroupIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("ReplaceMultipleHierarchies", string.Format("{0}-{1}", index, maxIndex));
            var replaceRequest = new replaceGroupsRequest { groupIdPairSet = objectSet.ToArray() };
            replaceGroupsResponse replaceResponse;

            // Sends the updated hierarchies to Itslearning.
            var response = _groupService.replaceGroups(headerInfo, replaceRequest, out replaceResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Deletes the given Hierarchies from Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="deleteSyncIds">The Hierarchies to be deleted.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        public BatchOperationResult DeleteHierarchies(ICollection<string> deleteSyncIds)
        {
            var maxCount = deleteSyncIds.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<string>(MaxOperations);

            foreach (var syncId in deleteSyncIds)
            {
                objectSet.Add(syncId);
                syncIds.Add(syncId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    DeleteHierarchies(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                DeleteHierarchies(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a DeleteHierarchies Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The hierarchies.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void DeleteHierarchies(List<string> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("DeleteMultipleHierarchies", string.Format("{0}-{1}", index, maxIndex));
            var deleteRequest = new deleteGroupsRequest { sourcedIdSet = objectSet.ToArray() };
            deleteGroupsResponse deleteResponse;

            // Deletes the hierarchies from Itslearning.
            var response = _groupService.deleteGroups(headerInfo, deleteRequest, out deleteResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Changes the SyncIds for the given Hierarchies.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="syncIdPairs">The old and new SyncIds.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        public BatchOperationResult ChangeHierarchyIds(ICollection<KeyPair> syncIdPairs)
        {
            var maxCount = syncIdPairs.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<IdentifierPairDType>(MaxOperations);

            foreach (var syncIdPair in syncIdPairs)
            {
                objectSet.Add(new IdentifierPairDType
                    {
                        firstId = syncIdPair.OldKey,
                        secondId = syncIdPair.NewKey
                    });
                syncIds.Add(syncIdPair.OldKey);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    ChangeHierarchyIds(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                ChangeHierarchyIds(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a ChangeHierarchyIds Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The hierarchies.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void ChangeHierarchyIds(List<IdentifierPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("ChangeMultipleHierarchyId", string.Format("{0}-{1}", index, maxIndex));
            var changeRequest = new changeGroupsIdentifierRequest { pairSourcedIdSet = objectSet.ToArray() };
            changeGroupsIdentifierResponse changeResponse;

            // Sends the new ids to Itslearning.
            var response = _groupService.changeGroupsIdentifier(headerInfo, changeRequest, out changeResponse);

            statusInfo.AddRange(response.statusInfoSet ?? new[] { response.statusInfo });
        }

        #endregion


        #region Course - Single Functions

        /// <summary>
        /// Creates a new Course in Itslearning.
        /// </summary>
        /// <param name="newCourse">The new Course.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the new Course is not valid.</exception>
        public OperationResult CreateCourse(Course newCourse)
        {
            var headerInfo = CreateHeader("CreateCourse", newCourse.SyncId);
            var createRequest = new createGroupRequest
            {
                sourcedId = new createGroupRequestSourcedId { identifier = newCourse.SyncId },
                group = newCourse.ToImsEsGroup(true)
            };
            createGroupResponse createResponse;

            // Sends the course to Itslearning.
            var response = _groupService.createGroup(headerInfo, createRequest, out createResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, newCourse.SyncId);
        }

        /// <summary>
        /// Retrieves a Course from Itslearning.
        /// </summary>
        /// <param name="syncId">The specified Course id.</param>
        /// <param name="readCourse">Out: the retrieved Course.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult ReadCourse(string syncId, out Course readCourse)
        {
            var headerInfo = CreateHeader("ReadCourse", syncId);
            var readRequest = new readGroupRequest
            {
                sourcedId = new readGroupRequestSourcedId { identifier = syncId },
            };
            readGroupResponse readResponse;

            // retrieves the course from Itslearning.
            var response = _groupService.readGroup(headerInfo, readRequest, out readResponse);
            readCourse = readResponse.group.ToItsCourse(syncId);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, syncId);
        }

        /// <summary>
        /// Updates the specified Course in Itslearning.
        /// </summary>
        /// <param name="updatedCourse">The updated course.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the updated Course is not valid.</exception>
        public OperationResult UpdateCourse(Course updatedCourse)
        {
            var headerInfo = CreateHeader("UpdateCourse", updatedCourse.SyncId);
            var updateRequest = new updateGroupRequest
                {
                    sourcedId = new updateGroupRequestSourcedId {identifier = updatedCourse.SyncId},
                    group = updatedCourse.ToImsEsGroup(false)
                };
            updateGroupResponse updateResponse;

            // Sends the updated course to Itslearning.
            var response = _groupService.updateGroup(headerInfo, updateRequest, out updateResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, updatedCourse.SyncId);
        }

        /// <summary>
        /// Replaces the specified Course in Itslearning (Update + all unspecified fields are emptied).
        /// </summary>
        /// <param name="updatedCourse">The updated course.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the replaced Course is not valid.</exception>
        public OperationResult ReplaceCourse(Course updatedCourse)
        {
            var headerInfo = CreateHeader("ReplaceCourse", updatedCourse.SyncId);
            var replaceRequest = new replaceGroupRequest
                {
                    sourcedId = new replaceGroupRequestSourcedId {identifier = updatedCourse.SyncId},
                    group = updatedCourse.ToImsEsGroup(true)
                };
            replaceGroupResponse replaceResponse;

            // Sends the updated course to Itslearning.
            var response = _groupService.replaceGroup(headerInfo, replaceRequest, out replaceResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, updatedCourse.SyncId);
        }

        /// <summary>
        /// Retrieves all Courses from Itslearning attached to the given User.
        /// </summary>
        /// <param name="syncKey">The User key.</param>
        /// <param name="readCourseList">Out: The list with Courses.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult ReadCoursesFromUser(string syncKey, out List<Course> readCourseList)
        {
            var headerInfo = CreateHeader("ReadCoursesFromUser", syncKey);
            var readRequest = new readGroupsForPersonRequest
            {
                personSourcedId = new readGroupsForPersonRequestPersonSourcedId { identifier = syncKey },
            };
            readGroupsForPersonResponse readResponse;

            // retrieves the courses from Itslearning.
            var response = _groupService.readGroupsForPerson(headerInfo, readRequest, out readResponse);
            readCourseList = readResponse.groupSet.ToItsCourseList();

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, syncKey);
        }

        #endregion

        
        #region Course - Batch Functions

        /// <summary>
        /// Creates the given Courses in Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="newCourses">The courses to be created.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If one of the new Courses is not valid.</exception>
        public BatchOperationResult CreateCourses(ICollection<Course> newCourses)
        {
            var maxCount = newCourses.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<GroupIdPairDType>(MaxOperations);

            foreach (var newCourse in newCourses)
            {
                objectSet.Add(new GroupIdPairDType
                    {
                        group = newCourse.ToImsEsGroup(true),
                        sourcedId = new GroupIdPairDTypeSourcedId{ identifier = newCourse.SyncId }
                    });
                syncIds.Add(newCourse.SyncId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    CreateCourses(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                CreateCourses(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a CreateCourses Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The courses.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void CreateCourses(List<GroupIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("CreateMultipleCourses", string.Format("{0}-{1}", index, maxIndex));
            var createRequest = new createGroupsRequest { groupIdPairSet = objectSet.ToArray() };
            createGroupsResponse createResponse;

            // Sends the courses to Itslearning.
            var response = _groupService.createGroups(headerInfo, createRequest, out createResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Retrieves the given Courses from Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="readSyncIds">The Courses to be retrieved.</param>
        /// <param name="readCourseList">Out: The retrieved courses.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        public BatchOperationResult ReadCourses(ICollection<string> readSyncIds, out List<Course> readCourseList)
        {
            var maxCount = readSyncIds.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            readCourseList = new List<Course>();
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<string>(MaxOperations);

            foreach (var syncId in readSyncIds)
            {
                objectSet.Add(syncId);
                syncIds.Add(syncId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    ReadCourses(objectSet, ref statusInfo, index, maxIndex, ref readCourseList);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                ReadCourses(objectSet, ref statusInfo, index, maxIndex, ref readCourseList);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a ReadCourses Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The courses.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        /// <param name="readCourseList">The list with retrieved Courses.</param>
        private void ReadCourses(List<string> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex, ref List<Course> readCourseList)
        {
            var headerInfo = CreateHeader("ReadMultipleCourses", string.Format("{0}-{1}", index, maxIndex));
            var readRequest = new readGroupsRequest { sourcedIdSet = objectSet.ToArray() };
            readGroupsResponse readResponse;

            // Retrieves the courses from Itslearning.
            var response = _groupService.readGroups(headerInfo, readRequest, out readResponse);
            readCourseList.AddRange(readResponse.groupSet.ToItsCourseList(objectSet, response));

            statusInfo.AddRange(response.statusInfoSet ?? new[] { response.statusInfo });
        }


        /// <summary>
        /// Updates the given Courses in Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="updatedCourses">The Courses to be updated.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If one of the updated Courses is not valid.</exception>
        public BatchOperationResult UpdateCourses(ICollection<Course> updatedCourses)
        {
            var maxCount = updatedCourses.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<GroupIdPairDType>(MaxOperations);

            foreach (var updatedCourse in updatedCourses)
            {
                objectSet.Add(new GroupIdPairDType
                    {
                        group = updatedCourse.ToImsEsGroup(false),
                        sourcedId = new GroupIdPairDTypeSourcedId {identifier = updatedCourse.SyncId}
                    });
                syncIds.Add(updatedCourse.SyncId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    UpdateCourses(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                UpdateCourses(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a CreateUsers Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The courses.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void UpdateCourses(List<GroupIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("UpdateMultipleCourses", string.Format("{0}-{1}", index, maxIndex));
            var updateRequest = new updateGroupsRequest { groupIdPairSet = objectSet.ToArray() };
            updateGroupsResponse updateResponse;

            // Sends the updated courses to Itslearning.
            var response = _groupService.updateGroups(headerInfo, updateRequest, out updateResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Replaces the given Courses in Itslearning (Update + all unspecified fields are emptied).
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="updatedCourses">The Courses to be updated.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If one of the replaced Courses is not valid.</exception>
        public BatchOperationResult ReplaceCourses(ICollection<Course> updatedCourses)
        {
            var maxCount = updatedCourses.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<GroupIdPairDType>(MaxOperations);

            foreach (var updatedCourse in updatedCourses)
            {
                objectSet.Add(new GroupIdPairDType
                    {
                        group = updatedCourse.ToImsEsGroup(true),
                        sourcedId = new GroupIdPairDTypeSourcedId {identifier = updatedCourse.SyncId}
                    });
                syncIds.Add(updatedCourse.SyncId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    ReplaceCourses(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                ReplaceCourses(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a ReplaceCourses Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The courses.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void ReplaceCourses(List<GroupIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("ReplaceMultipleCourses", string.Format("{0}-{1}", index, maxIndex));
            var replaceRequest = new replaceGroupsRequest { groupIdPairSet = objectSet.ToArray()};
            replaceGroupsResponse replaceResponse;

            // Sends the updated courses to Itslearning.
            var response = _groupService.replaceGroups(headerInfo, replaceRequest, out replaceResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }

        #endregion


        #region CourseGroup - Single Functions

        /// <summary>
        /// Creates a new CourseGroup in Itslearning.
        /// </summary>
        /// <param name="newCourseGroup">The new CourseGroup.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the new CourseGroup is not valid.</exception>
        public OperationResult CreateCourseGroup(CourseGroup newCourseGroup)
        {
            var headerInfo = CreateHeader("CreateCourseGroup", newCourseGroup.SyncId);
            var createRequest = new createGroupRequest
            {
                sourcedId = new createGroupRequestSourcedId { identifier = newCourseGroup.SyncId },
                group = newCourseGroup.ToImsEsGroup(true)
            };
            createGroupResponse createResponse;

            // Sends the CourseGroup to Itslearning.
            var response = _groupService.createGroup(headerInfo, createRequest, out createResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, newCourseGroup.SyncId);
        }

        /// <summary>
        /// Retrieves the specified CourseGroup from Itslearning.
        /// </summary>
        /// <param name="syncId">The specified CourseGroup Id.</param>
        /// <param name="readCourseGroup">Out: The retrieved CourseGroup.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult ReadCourseGroup(string syncId, out CourseGroup readCourseGroup)
        {
            var headerInfo = CreateHeader("ReadCourseGroup", syncId);
            var readRequest = new readGroupRequest
            {
                sourcedId = new readGroupRequestSourcedId { identifier = syncId },
            };
            readGroupResponse readResponse;

            // Retrieves the CourseGroup from Itslearning.
            var response = _groupService.readGroup(headerInfo, readRequest, out readResponse);
            readCourseGroup = readResponse.group.ToItsCourseGroup(syncId);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, syncId);
        }

        /// <summary>
        /// Updates an CourseGroup in Istlearning.
        /// </summary>
        /// <param name="updatedCourseGroup">The updated CourseGroup.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the updated CourseGroup is not valid.</exception>
        public OperationResult UpdateCourseGroup(CourseGroup updatedCourseGroup)
        {
            var headerInfo = CreateHeader("UpdateCourseGroup", updatedCourseGroup.SyncId);
            var updateRequest = new updateGroupRequest
            {
                sourcedId = new updateGroupRequestSourcedId { identifier = updatedCourseGroup.SyncId },
                group = updatedCourseGroup.ToImsEsGroup(false)
            };
            updateGroupResponse updateResponse;

            // Sends the updated CourseGroup to Itslearning.
            var response = _groupService.updateGroup(headerInfo, updateRequest, out updateResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, updatedCourseGroup.SyncId);
        }

        /// <summary>
        /// Replaces an CourseGroup in Istlearning (Update + all unspecified fields are emptied.).
        /// </summary>
        /// <param name="updatedCourseGroup">The updated CourseGroup.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the replaced CourseGroup is not valid.</exception>
        public OperationResult ReplaceCourseGroup(CourseGroup updatedCourseGroup)
        {
            var headerInfo = CreateHeader("ReplaceCourseGroup", updatedCourseGroup.SyncId);
            var replaceRequest = new replaceGroupRequest
            {
                sourcedId = new replaceGroupRequestSourcedId { identifier = updatedCourseGroup.SyncId },
                group = updatedCourseGroup.ToImsEsGroup(true)
            };
            replaceGroupResponse replaceResponse;

            // Sends the updated CourseGroup to Itslearning.
            var response = _groupService.replaceGroup(headerInfo, replaceRequest, out replaceResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, updatedCourseGroup.SyncId);
        }

        /// <summary>
        /// Deletes an CourseGroup in Itslearning.
        /// </summary>
        /// <param name="syncId">The id that must be deleted.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult DeleteCourseGroup(string syncId)
        {
            var headerInfo = CreateHeader("DeleteCourseGroup", syncId);
            var deleteRequest = new deleteGroupRequest
            {
                sourcedId = new deleteGroupRequestSourcedId { identifier = syncId },
            };
            deleteGroupResponse deleteResponse;

            // deletes the CourseGroup in Itslearning.
            var response = _groupService.deleteGroup(headerInfo, deleteRequest, out deleteResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, syncId);
        }

        /// <summary>
        /// Changes the id for the CourseGroup in Itslearning.
        /// </summary>
        /// <param name="oldId">The old Id.</param>
        /// <param name="newId">The new Id.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult ChangeCourseGroupId(string oldId, string newId)
        {
            var headerInfo = CreateHeader("ChangeCourseGroupId", oldId);
            var changeRequest = new changeGroupIdentifierRequest
            {
                sourcedId = new changeGroupIdentifierRequestSourcedId { identifier = oldId },
                newSourcedId = new changeGroupIdentifierRequestNewSourcedId { identifier = newId }
            };
            changeGroupIdentifierResponse changeResponse;

            // Sends the new id to Itslearning.
            var response = _groupService.changeGroupIdentifier(headerInfo, changeRequest, out changeResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, oldId);
        }

        /// <summary>
        /// Retrieves all CourseGroups from Itslearning attached to the given User.
        /// </summary>
        /// <param name="syncKey">The User.</param>
        /// <param name="readCourseGroupList">Out: The list with CourseGroup.</param>
        /// <returns>A (possible null) warning string.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult ReadCourseGroupsFromUser(string syncKey, out List<CourseGroup> readCourseGroupList)
        {
            var headerInfo = CreateHeader("ReadCourseGroupsFromUser", syncKey);
            var readRequest = new readGroupsForPersonRequest
            {
                personSourcedId = new readGroupsForPersonRequestPersonSourcedId { identifier = syncKey },
            };
            readGroupsForPersonResponse readResponse;

            // Retrieves the hierarchies from Itslearning.
            var response = _groupService.readGroupsForPerson(headerInfo, readRequest, out readResponse);
            readCourseGroupList = readResponse.groupSet.ToItsCourseGroupList();
            
            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, syncKey);
        }

        #endregion


        #region CourseGroup - Batch Functions

        /// <summary>
        /// Creates the given CourseGroups in Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="newCourseGroups">The new CourseGroups.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If one of the new CourseGroups is not valid.</exception>
        public BatchOperationResult CreateCourseGroups(ICollection<CourseGroup> newCourseGroups)
        {
            var maxCount = newCourseGroups.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<GroupIdPairDType>(MaxOperations);

            foreach (var newCourseGroup in newCourseGroups)
            {
                objectSet.Add(new GroupIdPairDType
                    {
                        group = newCourseGroup.ToImsEsGroup(true),
                        sourcedId = new GroupIdPairDTypeSourcedId {identifier = newCourseGroup.SyncId}
                    });
                syncIds.Add(newCourseGroup.SyncId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    CreateCourseGroups(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                CreateCourseGroups(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a CreateCourseGroups Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The course groups.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void CreateCourseGroups(List<GroupIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("CreateMultipleCourseGroups", string.Format("{0}-{1}", index, maxIndex));
            var createRequest = new createGroupsRequest { groupIdPairSet = objectSet.ToArray() };
            createGroupsResponse createResponse;

            // Sends the CourseGroups to Itslearning.
            var response = _groupService.createGroups(headerInfo, createRequest, out createResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Retrieves the given CourseGroups from Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="readSyncIds">The given CourseGroups.</param>
        /// <param name="readCourseGroupList">Out: The retrieved CourseGroups.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        public BatchOperationResult ReadCourseGroups(ICollection<string> readSyncIds, out List<CourseGroup> readCourseGroupList)
        {
            var maxCount = readSyncIds.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            readCourseGroupList = new List<CourseGroup>();
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<string>(MaxOperations);

            foreach (var syncId in readSyncIds)
            {
                objectSet.Add(syncId);
                syncIds.Add(syncId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    ReadCourseGroups(objectSet, ref statusInfo, index, maxIndex, ref readCourseGroupList);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
               ReadCourseGroups(objectSet, ref statusInfo, index, maxIndex, ref readCourseGroupList);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a ReadCourseGroups Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The course groups.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        /// <param name="readCourseGroupList">The list with retrieved CourseGroups.</param>
        private void ReadCourseGroups(List<string> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex, ref List<CourseGroup> readCourseGroupList)
        {
            var headerInfo = CreateHeader("ReadMultipleCourseGroups", string.Format("{0}-{1}", index, maxIndex));
            var readRequest = new readGroupsRequest { sourcedIdSet = objectSet.ToArray() };
            readGroupsResponse readResponse;

            // Retrieves the CourseGroups from Itslearning.
            var response = _groupService.readGroups(headerInfo, readRequest, out readResponse);
            readCourseGroupList.AddRange(readResponse.groupSet.ToItsCourseGroupList(objectSet, response));

            statusInfo.AddRange(response.statusInfoSet ?? new[] { response.statusInfo });
        }


        /// <summary>
        /// Updates the given CourseGroups in Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="updatedCourseGroups">The CourseGroups to be updated.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If one of the updated CourseGroups is not valid.</exception>
        public BatchOperationResult UpdateCourseGroups(ICollection<CourseGroup> updatedCourseGroups)
        {
            var maxCount = updatedCourseGroups.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<GroupIdPairDType>(MaxOperations);

            foreach (var updatedCourseGroup in updatedCourseGroups)
            {
                objectSet.Add(new GroupIdPairDType
                    {
                        group = updatedCourseGroup.ToImsEsGroup(false),
                        sourcedId = new GroupIdPairDTypeSourcedId {identifier = updatedCourseGroup.SyncId}
                    });
                syncIds.Add(updatedCourseGroup.SyncId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    UpdateCourseGroups(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                UpdateCourseGroups(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a UpdateCourseGroups Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The course groups.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void UpdateCourseGroups(List<GroupIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("UpdateMultipleCourseGroups", string.Format("{0}-{1}", index, maxIndex));
            var updateRequest = new updateGroupsRequest { groupIdPairSet = objectSet.ToArray() };
            updateGroupsResponse updateResponse;

            // Sends the updated CourseGroups to Itslearning.
            var response = _groupService.updateGroups(headerInfo, updateRequest, out updateResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Replaces the given CourseGroups in Itslearning (Update + any unspecified fields are emptied).
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="updatedCourseGroups">The CourseGroups to be updated.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If one of the replaced CourseGroups is not valid.</exception>
        public BatchOperationResult ReplaceCourseGroups(ICollection<CourseGroup> updatedCourseGroups)
        {
            var maxCount = updatedCourseGroups.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<GroupIdPairDType>(MaxOperations);

            foreach (var updatedCourseGroup in updatedCourseGroups)
            {
                objectSet.Add(new GroupIdPairDType
                    {
                        group = updatedCourseGroup.ToImsEsGroup(true),
                        sourcedId = new GroupIdPairDTypeSourcedId {identifier = updatedCourseGroup.SyncId}
                    });
                syncIds.Add(updatedCourseGroup.SyncId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    ReplaceCourseGroups(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                ReplaceCourseGroups(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a ReplaceCourseGroups Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The course groups.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void ReplaceCourseGroups(List<GroupIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("ReplaceMultipleCourseGroups", string.Format("{0}-{1}", index, maxIndex));
            var replaceRequest = new replaceGroupsRequest {groupIdPairSet = objectSet.ToArray()};
            replaceGroupsResponse replaceResponse;

            // Sends the updated CourseGroups to Itslearning.
            var response = _groupService.replaceGroups(headerInfo, replaceRequest, out replaceResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Deletes the given CourseGroups from Itslearning.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="deleteSyncIds">The CourseGroups to be deleted.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        public BatchOperationResult DeleteCourseGroups(ICollection<string> deleteSyncIds)
        {
            var maxCount = deleteSyncIds.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<string>(MaxOperations);

            foreach (var syncId in deleteSyncIds)
            {
                objectSet.Add(syncId);
                syncIds.Add(syncId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    DeleteCourseGroups(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                DeleteCourseGroups(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a DeleteCourseGroups Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The course groups.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void DeleteCourseGroups(List<string> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("DeleteMultipleCourseGroups", string.Format("{0}-{1}", index, maxIndex));
            var deleteRequest = new deleteGroupsRequest { sourcedIdSet = objectSet.ToArray() };
            deleteGroupsResponse deleteResponse;

            // Deletes the CourseGroups from Itslearning.
            var response = _groupService.deleteGroups(headerInfo, deleteRequest, out deleteResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Changes the SyncIds for the given CourseGroups.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="syncIdPairs">The old and new SyncIds.</param>
        /// <returns>A (possible empty) List with warning messages.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        public BatchOperationResult ChangeCourseGroupIds(ICollection<KeyPair> syncIdPairs)
        {
            var maxCount = syncIdPairs.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<IdentifierPairDType>(MaxOperations);

            foreach (var syncIdPair in syncIdPairs)
            {
                objectSet.Add(new IdentifierPairDType
                {
                    firstId = syncIdPair.OldKey,
                    secondId = syncIdPair.NewKey
                });
                syncIds.Add(syncIdPair.OldKey);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    ChangeCourseGroupIds(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                ChangeCourseGroupIds(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a ChangeCourseGroupIds Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The course groups.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void ChangeCourseGroupIds(List<IdentifierPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("ChangeMultipleCourseGroupIds", string.Format("{0}-{1}", index, maxIndex));
            var changeRequest = new changeGroupsIdentifierRequest { pairSourcedIdSet = objectSet.ToArray() };
            changeGroupsIdentifierResponse changeResponse;

            // Sends the new ids to Itslearning.
            var response = _groupService.changeGroupsIdentifier(headerInfo, changeRequest, out changeResponse);

            statusInfo.AddRange(response.statusInfoSet ?? new[] { response.statusInfo });
        }

        #endregion


        #region Membership - Single Functions

        /// <summary>
        /// Creates a new MembershipId, based on the UserKey and HierarchyId.
        /// </summary>
        /// <param name="userSyncKey">The User SyncKey.</param>
        /// <param name="hierarchySyncId">The Hierarchy SyncId.</param>
        /// <returns>The new MembershipId.</returns>
        /// <exception cref="ValidationException">If the total length > 255 characters.</exception>
        private static string GetMembershipId(string userSyncKey, string hierarchySyncId)
        {
            var membershipId = string.Format("imses_automated_{0}-{1}", userSyncKey, hierarchySyncId);
            if (membershipId.Length > 255)
            {
                throw new ValidationException("The length of the UserKey and HierarchyId is too long, please specify a shorter Membership Id.");
            }
            return membershipId;
        }


        /// <summary>
        /// Creates a membership that Adds the given user to the given Hierarchy, with the specified Role.
        /// (Note: the length of the UserKey + the HierarchyId can be at most 235 characters!)
        /// </summary>
        /// <param name="userSyncKey">The User to be added.</param>
        /// <param name="hierarchySyncId">Hierarchy.</param>
        /// <param name="role">Role of the User.</param>
        /// <returns>A (possible null) warning.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the length of the membershipId > 255 characters.</exception>
        public OperationResult CreateMembership(string userSyncKey, string hierarchySyncId, HierarchyRole role)
        {
            var membershipId = GetMembershipId(userSyncKey, hierarchySyncId);
            return CreateMembership(membershipId, userSyncKey, hierarchySyncId, role);
        }

        /// <summary>
        /// Creates a membership that Adds the given user to the given Hierarchy, with the specified Role.
        /// (Note: the length of the UserKey + the HierarchyId can be at most 235 characters!)
        /// </summary>
        /// <param name="userSyncKey">The User to be added.</param>
        /// <param name="hierarchySyncId">Hierarchy.</param>
        /// <param name="role">Role of the User.</param>
        /// <param name="membershipId">The new Membership Id.</param>
        /// <returns>A (possible null) warning.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the length of the membershipId > 255 characters.</exception>
        public OperationResult CreateMembership(string userSyncKey, string hierarchySyncId, HierarchyRole role, out string membershipId)
        {
            membershipId = GetMembershipId(userSyncKey, hierarchySyncId);
            return CreateMembership(membershipId, userSyncKey, hierarchySyncId, role);
        }

        /// <summary>
        /// Creates a membership that Adds the given user to the given Hierarchy, with the specified Role.
        /// (Note: the length of the MembershipId can be at most 255 characters!)
        /// </summary>
        /// <param name="membershipId">The new Membership Id.</param>
        /// <param name="userSyncKey">The User to be added.</param>
        /// <param name="hierarchySyncId">Hierarchy.</param>
        /// <param name="role">Role of the User.</param>
        /// <returns>A (possible null) warning.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult CreateMembership(string membershipId, string userSyncKey, string hierarchySyncId, HierarchyRole role)
        {
            var headerInfo = CreateHeader("CreateMembership", membershipId);
            var createRequest = new createMembershipRequest
                {
                    sourcedId = new createMembershipRequestSourcedId {identifier = membershipId},
                    membership = ConvertObjects.ToImsEsMembership(userSyncKey, hierarchySyncId, role)
                };
            createMembershipResponse createResponse;

            // Send the membership to Itslearning.
            var response = _memberService.createMembership(headerInfo, createRequest, out createResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, membershipId);
        }


        /// <summary>
        /// Updates the Membership, by changing the role of the User in the Hierarchy.
        /// (note: if at the moment this membership was created a different MembershipId was given,
        /// this function will not work!)
        /// </summary>
        /// <param name="userSyncKey">The User.</param>
        /// <param name="hierarchySyncId">The Hierarchy.</param>
        /// <param name="role">The Role.</param>
        /// <returns>A (possible null) warning.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the length of the membershipId > 255 characters.</exception>
        public OperationResult UpdateMembership(string userSyncKey, string hierarchySyncId, HierarchyRole role)
        {
            var membershipId = GetMembershipId(userSyncKey, hierarchySyncId);
            return UpdateMembership(membershipId, userSyncKey, hierarchySyncId, role);
        }

        /// <summary>
        /// Updates the given Membership, by changing the role of the User in the Hierarchy.
        /// </summary>
        /// <param name="membershipId">The Membership.</param>
        /// <param name="userSyncKey">The User.</param>
        /// <param name="hierarchySyncId">The Hierarchy.</param>
        /// <param name="role">The Role.</param>
        /// <returns>A (possible null) warning.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult UpdateMembership(string membershipId, string userSyncKey, string hierarchySyncId, HierarchyRole role)
        {
            var headerInfo = CreateHeader("UpdateMembership", membershipId);
            var updateRequest = new updateMembershipRequest
                {
                    sourcedId = new updateMembershipRequestSourcedId {identifier = membershipId},
                    membership = ConvertObjects.ToImsEsMembership(userSyncKey, hierarchySyncId, role)
                };
            updateMembershipResponse updateResponse;

            // Send the membership to Itslearning.
            var response = _memberService.updateMembership(headerInfo, updateRequest, out updateResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, membershipId);
        }


        /// <summary>
        /// Removes the User from the Hierarchy, by deleting the Memberhsip.
        /// (note: if at the moment this membership was created a different MembershipId was given,
        /// this function will not work!)
        /// </summary>
        /// <param name="userSyncKey">The User.</param>
        /// <param name="hierarchySyncId">The Hierarchy.</param>
        /// <returns>A (possible null) warning.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the length of the membershipId > 255 characters.</exception>
        public OperationResult DeleteMembership(string userSyncKey, string hierarchySyncId)
        {
            var membershipId = GetMembershipId(userSyncKey, hierarchySyncId);
            return DeleteMembership(membershipId);
        }

        /// <summary>
        /// Removes the User from the Hierarchy, by deleting the given Memberhsip.
        /// </summary>
        /// <param name="membershipId">The Membership.</param>
        /// <returns>A (possible null) warning.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult DeleteMembership(string membershipId)
        {
            var headerInfo = CreateHeader("DeleteMembership", membershipId);
            var deleteRequest = new deleteMembershipRequest
                {
                    sourcedId = new deleteMembershipRequestSourcedId {identifier = membershipId}
                };
            deleteMembershipResponse deleteResponse;

            // Send the membership to Itslearning.
            var response = _memberService.deleteMembership(headerInfo, deleteRequest, out deleteResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, membershipId);
        }
        

        /// <summary>
        /// Sets the Hierarchy as the home organisation of the User.
        /// (note: if at the moment this membership was created a different MembershipId was given,
        /// this function will not work!)
        /// </summary>
        /// <param name="userSyncKey">The User.</param>
        /// <param name="hierarchySyncId">The Hierarchy.</param>
        /// <returns>A (possible null) warning.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the length of the membershipId > 255 characters.</exception>
        public OperationResult SetHomeOrganisation(string userSyncKey, string hierarchySyncId)
        {
            var memberhsipId = GetMembershipId(userSyncKey, hierarchySyncId);
            return SetHomeOrganisation(memberhsipId, userSyncKey, hierarchySyncId);
        }

        /// <summary>
        /// Sets the Hierarchy as the home organisation of the User.
        /// </summary>
        /// <param name="membershipId">The Membership.</param>
        /// <param name="userSyncKey">The User.</param>
        /// <param name="hierarchySyncId">The Hierarchy.</param>
        /// <returns>A (possible null) warning.</returns>
        /// <exception cref="ImsEsException">If ImsEs returns an error.</exception>
        public OperationResult SetHomeOrganisation(string membershipId, string userSyncKey, string hierarchySyncId)
        {
            var headerInfo = CreateHeader("SetHomeOrganisation", membershipId);
            var updateRequest = new updateMembershipRequest
            {
                sourcedId = new updateMembershipRequestSourcedId { identifier = membershipId },
                membership = ConvertObjects.ToImsEsHomeMembership(userSyncKey, hierarchySyncId)
            };
            updateMembershipResponse updateResponse;

            // Send the membership to Itslearning.
            var response = _memberService.updateMembership(headerInfo, updateRequest, out updateResponse);

            // Check for exceptions and warnings.
            return CheckSingleErrorWarning(response, membershipId);
        }
        
        #endregion

        
        #region Membership - Batch Functions

        /// <summary>
        /// Creates multiple memberships that Adds the given users to the given Hierarchys, with the specified Roles.
        /// (Note: the length of the UserKey + the HierarchyId can be at most 235 characters!)
        /// (Also: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="memberships">The User / Hierarchy Pairs, with the Role of the User.</param>
        /// <returns>A (possible empty) list with warning.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the length of one of the membershipIds > 255 characters.</exception>
        public BatchOperationResult CreateMemberships(ICollection<MembershipPair> memberships)
        {
            var dictionary = memberships.ToDictionary(pair => GetMembershipId(pair.UserSyncKey, pair.HierarchySyncId));
            return CreateMemberships(dictionary);
        }

        /// <summary>
        /// Creates multiple memberships that foe every pair Adds the given user to the given Hierarchy, with the specified Role.
        /// (Note: the length of the UserKey + the HierarchyId can be at most 235 characters!)
        /// (Also: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="memberships">The User / Hierarchy Pairs, with the Role of the User.</param>
        /// <param name="membershipIds">Out: The generated MembershipId.</param>
        /// <returns>A (possible empty) list with warning.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the length of one of the membershipIds > 255 characters.</exception>
        public BatchOperationResult CreateMemberships(ICollection<MembershipPair> memberships, out List<string> membershipIds)
        {
            var dictionary = memberships.ToDictionary(pair => GetMembershipId(pair.UserSyncKey, pair.HierarchySyncId));
            membershipIds = dictionary.Keys.ToList();
            return CreateMemberships(dictionary);
        }

        /// <summary>
        /// Creates multiple memberships that foe every pair Adds the given user to the given Hierarchy, with the specified Role,
        /// using the specified MembershipId.
        /// (Note: the length of the MembershipId can be at most 255 characters!)
        /// (Also: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="memberships">The User / Hierarchy Pairs, with the Role of the User.</param>
        /// <returns>A (possible empty) list with warning.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        public BatchOperationResult CreateMemberships(IDictionary<string, MembershipPair> memberships)
        {
            var maxCount = memberships.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<MembershipIdPairDType>(MaxOperations);

            foreach (var membership in memberships)
            {
                objectSet.Add(new MembershipIdPairDType
                    {
                        sourcedId = new MembershipIdPairDTypeSourcedId { identifier = membership.Key },
                        membership = membership.Value.ToImsEsMembership()
                    });
                syncIds.Add(membership.Key);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    CreateMemberships(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                CreateMemberships(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a CreateMemberships Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The memberships.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void CreateMemberships(List<MembershipIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("CreateMultipleMemberships", string.Format("{0}-{1}", index, maxIndex));
            var createRequest = new createMembershipsRequest { membershipIdPairSet = objectSet.ToArray() };
            createMembershipsResponse createResponse;

            // Sends the memberships to Itslearning.
            var response = _memberService.createMemberships(headerInfo, createRequest, out createResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Updates multiple Memberships, by changing the role of the Users in the Hierarchys.
        /// (Note: if at the moment where one of the memberships was created a different MembershipId was given,
        /// this function will not work!)
        /// (Also: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="memberships">The User / Hierarchy Pairs, with the Role of the User.</param>
        /// <returns>A (possible empty) list with warning.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the length of one of the membershipIds > 255 characters.</exception>
        public BatchOperationResult UpdateMemberships(ICollection<MembershipPair> memberships)
        {
            var dictionary = memberships.ToDictionary(pair => GetMembershipId(pair.UserSyncKey, pair.HierarchySyncId));
            return UpdateMemberships(dictionary);
        }

        /// <summary>
        /// Updates multiple Memberships, by changing the role of the Users in the Hierarchys.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="memberships">The User / Hierarchy Pairs, with the Role of the User.</param>
        /// <returns>A (possible empty) list with warning.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        public BatchOperationResult UpdateMemberships(IDictionary<string, MembershipPair> memberships)
        {
            var maxCount = memberships.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<MembershipIdPairDType>(MaxOperations);

            foreach (var membership in memberships)
            {
                objectSet.Add(new MembershipIdPairDType
                    {
                        sourcedId = new MembershipIdPairDTypeSourcedId {identifier = membership.Key},
                        membership = membership.Value.ToImsEsMembership()
                    });
                syncIds.Add(membership.Key);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    UpdateMemberships(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                UpdateMemberships(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a UpdateMemberships Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The memberships.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void UpdateMemberships(List<MembershipIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("UpdateMultipleMemberships", string.Format("{0}-{1}", index, maxIndex));
            var updateRequest = new updateMembershipsRequest { membershipIdPairSet = objectSet.ToArray() };
            updateMembershipsResponse updateResponse;

            // Sends the updated memberships to Itslearning.
            var response = _memberService.updateMemberships(headerInfo, updateRequest, out updateResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// For every Pair removes the User from the Hierarchy, by deleting the Membership.
        /// (Note: the role doesn't have to be filled in. Also: if at the moment one of the memberships were created a 
        /// different MembershipId was given, this function will not work!)
        /// (Also: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="memberships">The Membership Pairs.</param>
        /// <returns>A (possible empty) list with warning.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the length of any of the membershipIds > 255 characters.</exception>
        public BatchOperationResult DeleteMemberships(ICollection<MembershipPair> memberships)
        {
            var membershipIds = memberships.Select(x => GetMembershipId(x.UserSyncKey, x.HierarchySyncId)).ToArray();
            return DeleteMemberships(membershipIds);
        }

        /// <summary>
        /// For every membership removes the User from the Hierarchy, by deleting the Membership.
        /// (Note: the role doesn't have to be filled in.)
        /// (Also: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="membershipIds">The Memberships.</param>
        /// <returns>A (possible empty) list with warning.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        public BatchOperationResult DeleteMemberships(ICollection<string> membershipIds)
        {
            var maxCount = membershipIds.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<string>(MaxOperations);

            foreach (var membershipId in membershipIds)
            {
                objectSet.Add(membershipId);
                syncIds.Add(membershipId);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    DeleteMemberships(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                DeleteMemberships(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a DeleteMemberships Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The memberships.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        /// <returns>The returned status info.</returns>
        private void DeleteMemberships(List<string> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("DeleteMultipleMemberships", string.Format("{0}-{1}", index, maxIndex));
            var deleteRequest = new deleteMembershipsRequest { sourcedIdSet = objectSet.ToArray() };
            deleteMembershipsResponse deleteResponse;

            // Deletes the memberships from Itslearning.
            var response = _memberService.deleteMemberships(headerInfo, deleteRequest, out deleteResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }


        /// <summary>
        /// Sets for every pair the Hierarchy as the home organisation of the User.
        /// (Note: if at the moment one of these memberships were created a different MembershipId was given,
        /// this function will not work!)
        /// (Also: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="memberships">The membership pairs.</param>
        /// <returns>A (possible empty) list with warning.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        /// <exception cref="ValidationException">If the length of on of the membershipIds > 255 characters.</exception>
        public BatchOperationResult SetHomeOrganisations(ICollection<MembershipPair> memberships)
        {
            var dictionary = memberships.ToDictionary(pair => GetMembershipId(pair.UserSyncKey, pair.HierarchySyncId));
            return SetHomeOrganisations(dictionary);
        }

        /// <summary>
        /// Sets for every pair the Hierarchy as the home organisation of the User.
        /// (Note: if there are too many objects, the request will be split into several parts)
        /// </summary>
        /// <param name="memberships">The memberships.</param>
        /// <returns>A (possible empty) list with warning.</returns>
        /// <exception cref="ImsEsBatchException">If ImsEs returns an error.</exception>
        public BatchOperationResult SetHomeOrganisations(IDictionary<string, MembershipPair> memberships)
        {
            var maxCount = memberships.Count;
            var index = 1;
            var maxIndex = (int)Math.Ceiling((double)maxCount / MaxOperations);

            // Create some Lists for handling all the objects.
            var statusInfo = new List<StatusInfoDType>(maxCount);
            var syncIds = new List<string>(maxCount);
            var objectSet = new List<MembershipIdPairDType>(MaxOperations);

            foreach (var membership in memberships)
            {
                objectSet.Add(new MembershipIdPairDType
                {
                    sourcedId = new MembershipIdPairDTypeSourcedId { identifier = membership.Key },
                    membership = membership.Value.ToImsEsHomeMembership()
                });
                syncIds.Add(membership.Key);

                // If we have enough objects we can send them to Itslearning.
                if (objectSet.Count >= MaxOperations)
                {
                    SetHomeOrganisations(objectSet, ref statusInfo, index, maxIndex);

                    objectSet.Clear();
                    index++;
                }
            }
            // And send of the last batch of users.
            if (objectSet.Count > 0)
            {
                SetHomeOrganisations(objectSet, ref statusInfo, index, maxIndex);
            }

            // Check for exceptions and warnings.
            return CheckMultipleErrorWarning(statusInfo, syncIds);
        }

        /// <summary>
        /// Private function to send a SetHomeOrganisations Batch to Itslearning.
        /// </summary>
        /// <param name="objectSet">The memberships.</param>
        /// <param name="statusInfo">StatusInfo.</param>
        /// <param name="index">index of the request.</param>
        /// <param name="maxIndex">total number of requests.</param>
        private void SetHomeOrganisations(List<MembershipIdPairDType> objectSet, ref List<StatusInfoDType> statusInfo, int index, int maxIndex)
        {
            var headerInfo = CreateHeader("SetMultipleHomeOrganisations", string.Format("{0}-{1}", index, maxIndex));
            var updateRequest = new updateMembershipsRequest { membershipIdPairSet = objectSet.ToArray() };
            updateMembershipsResponse updateResponse;

            // Sends the updated memberships to Itslearning.
            var response = _memberService.updateMemberships(headerInfo, updateRequest, out updateResponse);

            statusInfo.AddRange(response.statusInfoSet);
        }

        #endregion

    }
}
