﻿using System.ComponentModel.DataAnnotations;
using System.Data;

namespace IMSES_Template.Model
{
    public class Course : BaseEntity
    {
        /// <summary>
        /// The SyncId of the Course (also known as Synchonisation ID).
        /// </summary>
        [Required]
        public string SyncId { get; protected set; }

        /// <summary>
        /// The Title of the Course.
        /// </summary>
        [Required]
        public string Title { get; set; }

        // Note: when updating a course, if you don't specify the Title, the Code and Credit will not be updated!

        /// <summary>
        /// The Course Code.
        /// </summary>
        public string Code { get; set; }

        /// <summary>
        /// The Credit Points for this Course (Also known as Hours).
        /// </summary>
        public string Credit { get; set; }

        /// <summary>
        /// Private: the Linked Hierarchy.
        /// </summary>
        [Required] 
        private Hierarchy _linkedHierarchy;

        /// <summary>
        /// The linked Hierarchy in Itslearning.
        /// </summary>
        public Hierarchy LinkedHierarchy
        {
            get { return _linkedHierarchy; }
            set
            {
                if (value == null)
                {
                    throw new NoNullAllowedException("Linked Hierarchy cannot be null!");
                }
                if (!value.SyncId.Equals(SyncId))
                {
                    throw new ValidationException("The Linked Hierarchy does not have the same SyncId as the Course!");
                }

                _linkedHierarchy = value;
                _linkedHierarchy.AttachedCourse = this;
            }
        }

        
        /// <summary>
        /// When creating a Course object the SyncId is required.
        /// </summary>
        /// <param name="syncId">The SyncId of the Course.</param>
        public Course(string syncId)
        {
            SyncId = syncId;
        }

        /// <summary>
        /// Creates a new Course based on the linked Hierarchy.
        /// </summary>
        /// <param name="linkedHierarchy">The linked Hierarchy.</param>
        public Course(Hierarchy linkedHierarchy)
            : this(linkedHierarchy.SyncId, linkedHierarchy)
        {
        }

        /// <summary>
        /// Creates a new Course with the given SyncId and linked hierarchy.
        /// </summary>
        /// <param name="syncId"></param>
        /// <param name="linkedHierarchy"></param>
        public Course(string syncId, Hierarchy linkedHierarchy)
        {
            SyncId = syncId;
            LinkedHierarchy = linkedHierarchy;
        }


        /// <summary>
        /// Checks if the object is valid, i.e. that all properties marked with [Required] are actually initialized and filled, 
        /// and throws an exception with the porperties that aren't. It also checks the Linked Hierarchy.
        /// </summary>
        public void CheckRequired()
        {
            CheckRequired(SyncId);

            if (LinkedHierarchy != null)
            {
                LinkedHierarchy.CheckSingleRequired();
            }
        }

        /// <summary>
        /// Checks if the object is valid, i.e. that all properties marked with [Required] are actually initialized and filled, 
        /// and throws an exception with the porperties that aren't.
        /// </summary>
        internal void CheckSingleRequired()
        {
            CheckRequired(SyncId);
        }


        /// <summary>
        /// Checks if all required fields are filled in, and checks if certain properties are not too long. It also checks the Linked Hierarchy.
        /// </summary>
        /// <returns>Either the properties that are required and not filled in, or an error about the length of an property.</returns>
        public void CheckValid()
        {
            CheckSingleValid();

            if (LinkedHierarchy != null)
                LinkedHierarchy.CheckSingleValid();
        }

        /// <summary>
        /// Checks if all required fields are filled in, and checks if certain properties are not too long.
        /// </summary>
        /// <returns>Either the properties that are required and not filled in, or an error about the length of an property.</returns>
        internal void CheckSingleValid()
        {
            // Check if certain properties are not too long.
            if (SyncId.Length > 255)
                throw new ValidationException(string.Format("({0}) SyncId is too long, max length is 255 characters.", SyncId));

            if (Title != null && Title.Length > 64)
                throw new ValidationException(string.Format("({0}) Title is too long, max length is 64 characters.", SyncId));

            if (Code != null && Code.Length > 64)
                throw new ValidationException(string.Format("({0}) Code is too long, max length is 64 characters.", SyncId));

            if (Credit != null && Credit.Length > 64)
                throw new ValidationException(string.Format("({0}) Credit is too long, max length is 64 characters.", SyncId));
        }

    }
}
